! Module      : omcosflag_mod
!
! Authors     : Tim Harries (MSSL)
!             : Phil Williams (MSSL)
!             : Andrew King (MSSL)
!             : Chris Brindle (MSSL)
! Description : Applies shift and add to bad pixel map.
!
! Comments    : 
!
! $Id: module.f90,v 1.4 1997/12/18 16:01:10 gvacanti Exp $
!
! History : Created  : January 98 - TJH
!           Modified : 27 April 98 - Removed twixel workaround.
!                                  - Removed datestamp call.
!                      20 March 99 - Spring clean of code.
!                      14 May 99 - Update shift and add process.
!                      
!
!                      24/08/2000 - AMK
!                      Initialised qualArraP
!
!                      24/08/2000 - AMK
!                      Changed incorrect initialisation of quality array.
!                      integer array had been initliased to real values
!
!                      Version 1.17.3: 14th June 2000
!                      Introduced new key word deAllocationError for warnings.
!
!                      Version 1.17.1: 9th June 2000
!                      Removed warning for bad pixel information used as test.
!                      We believe that the call to get` the bad pixel map works
!                      We do not know if it is correct

!                      Cleaned up memory allocation

!                      Removed forcing of one bad pixel - hardwired by JFR
!                      presumably for testing
!
!                      Version 1.15: 11th April 2000
!                      Modified 'zero drift' algorithm 		
!
!                      Version 1.14: 2nd March 2000
!                      Modified algorithm to cope with no drift file	
!	
!                      Version 1.13: 16th December 1999
!                      Updated test harness wrt new CCF for bad pixels
!                      Added new parameter timeSample	
!	
!                      Version 1.12: 2nd December 1999
!                      Added DISTRIBUTION (world) file	
!	
!                      Version 1.11: 8th June 1999
!                      Updated dependancies for V3.
!
!                      Version 1.10: 18th May 1999
!                      Removed CAL call to open CCF.
!                      Changed fatal datamode message.
!                      12.10.2001 - Program tidied up
!////////////////////////////////////////////////////////////////////
! 
! PATCHES AND FIXES AREA
!
! 1. 23.05.2000 - Removed forcing of bad pixels - believe that getBadPixelMap is working
!
!////////////////////////////////////////////////////////////////////
module omcosflag_mod

  use param
  use cal
  use dal
  use errorHandling
  use readHist_mod
  use readModes_mod

  implicit none

  private

  public omcosflag

contains

subroutine omcosflag
    
    type(DataSetT) :: inDataSetH                                      ! Input dataset handle.
    type(DataSetT) :: thxDataSetH                                     ! Tracking History Data Auxiliary dataset handle.
    type(TableT) :: inTableH                                          ! Input table handle.
    type(TableT) :: thxTableH                                         ! Tracking History Data Auxiliary table handle.
    type(ColumnT), dimension(4) :: inColumnH                          ! Input image column handle.
    type(ColumnT), dimension(3) :: thxColumnH                         ! Tracking History Data Auxiliary column handle.
    type(ArrayT) :: qualArrayH                                        ! Quality Array handle.
    type(ArrayT) :: imageArrayH                                       ! Image array handle
    real(kind=single), dimension(:, :), pointer :: imageArrayP        ! Pointer to image array

    type trackingFrame
       real(kind=single) :: dX                                        ! Delta x.
       real(kind=single) :: dY                                        ! Delta y.
       real(kind=single) :: quality                                   ! Quality.
    end type trackingFrame
    type(trackingFrame), allocatable, dimension(:) :: tf              ! Tracking frame array.


    character(len=8) :: dataMode                                      ! Input file data mode.
    character(len=255) :: set                                         ! OM OSW Image filename.
    character(len=255) :: thxSet                                      ! OM Tracking History Data Auxiliary filename.

    integer(kind=int8), dimension(:, :), pointer :: qualArrayP        ! Quality array pointer.

    integer(kind=int16), dimension(:, :), pointer :: badPixMapP       ! Bad pixel map pointer.
    integer(kind=int16), dimension(:, :), allocatable :: windowedbadPixMap ! Bad pixel map.
    integer(kind=int16), dimension(:, :), allocatable :: binnedbadPixMap ! Bad pixel map.
    integer(kind=int16) :: nFrames                                    ! Number of tracking frames. 

    integer(kind=int32) :: sampleFactor                               ! Sampling factor.
    integer(kind=int32) :: timeFactor                                 ! Time sampling factor
    integer(kind=int32), dimension(:), pointer :: windowX0P           ! WindowX0 pointer.
    integer(kind=int32), dimension(:), pointer :: windowY0P           ! WindowY0 pointer.
    integer(kind=int32), dimension(:), pointer :: windowDXP           ! WindowDX pointer.
    integer(kind=int32), dimension(:), pointer :: windowDYP           ! WindowDY pointer.
    integer(kind=int32), dimension(:), pointer :: dXP                 ! DX pointer.
    integer(kind=int32), dimension(:), pointer :: dYP                 ! DY pointer.
    integer(kind=int32), dimension(:), pointer :: qualityP            ! Quality pointer.
    integer(kind=int32) :: binAx1                                     ! DPU x-axis binning factors.
    integer(kind=int32) :: binAx2                                     ! DPU y-axis binning factors.
    integer(kind=int32) :: windowX0                                   ! WindowX0.
    integer(kind=int32) :: windowY0                                   ! WindowY0.
    integer(kind=int32) :: windowDX                                   ! WindowDX.
    integer(kind=int32) :: windowDY                                   ! WindowDY.
    integer(kind=int32) :: allocErr                                   ! Allocate error status.
    integer(kind=int32) :: maxDX                                      ! Maximum drift in x (centroided pixels).
    integer(kind=int32) :: maxDY                                      ! Maximum drift in y (centroided pixels).
    integer(kind=int32) :: cpxMin                                     ! Minimum x value for window (centroided pixels).
    integer(kind=int32) :: cpxMax                                     ! Maximum x value for window (centroided pixels).
    integer(kind=int32) :: cpyMin                                     ! Minimum y value for window (centroided pixels).
    integer(kind=int32) :: cpyMax                                     ! Maximum y value for window (centroided pixels).
    integer(kind=int32) :: i                                          ! Count.
    integer(kind=int32) :: j                                          ! Count.
    integer(kind=int32) :: oMaxDX                                     ! Maximum drift in x (oversampled pixels).
    integer(kind=int32) :: oMaxDY                                     ! Maximum drift in y (oversampled pixels).
    integer(kind=int32) :: opxMin                                     ! Minimum x value for window (oversampled pixels).
    integer(kind=int32) :: opxMax                                     ! Maximum x value for window (oversampled pixels).
    integer(kind=int32) :: opyMin                                     ! Minimum y value for window (oversampled pixels).
    integer(kind=int32) :: opyMax                                     ! Maximum y value for window (oversampled pixels).
    integer(kind=int32) :: shiftX                                     ! Shift in x.
    integer(kind=int32) :: shiftY                                     ! Shift in y.
    integer(kind=int32) :: x                                          ! X.
    integer(kind=int32) :: y                                          ! Y.
    integer(kind=int32) :: iShiftX                                    ! Incremental shift in x.
    integer(kind=int32) :: iShiftY                                    ! Incremental shift in y.
    integer(kind=int32) :: iXMin                                      ! Incremental minimum x value for window.
    integer(kind=int32) :: iXMax                                      ! Incremental maximum x value for window.
    integer(kind=int32) :: iYMin                                      ! Incremental minimum y value for window.
    integer(kind=int32) :: iYMax                                      ! Incremental maximum y value for window.
    integer(kind=int32) :: xConstant, yConstant
    integer(kind=int32), dimension(2) :: qualArrayDim                 ! Quality array dimension.
    integer(kind=int32) :: binX                                       ! Bin x.
    integer(kind=int32) :: binY                                       ! Bin y.
    integer(kind=int32) :: iPix, jPix
    integer(kind=int32), dimension(:), allocatable :: yIndex          ! Stores j array-element indices               
    real(kind=single) :: rShiftX                                      ! Relative shift in x.
    real(kind=single) :: rShiftY                                      ! Relative shift in y.
    real(kind=single) :: missingPixelValue
    logical(kind=bool) :: binBPE                                      ! BPE binning flag.
    logical(kind=bool), dimension(:, :), allocatable :: badPixMap     ! Bad pixel map.
    logical(kind=bool), dimension(:, :), allocatable :: lBadPixMap    ! Localised bad pixel map.
    logical(kind=bool), dimension(:, :), allocatable :: oBadPixMap    ! Oversampled bad pixel map.
    logical(kind=bool), dimension(:, :), allocatable :: oQualArray    ! Oversampled quality array.
    logical(kind=bool), dimension(:, :), allocatable :: qualArray     ! Quality array.
    integer iMax, jMax
    integer            :: istat    ! Status flag
    integer, dimension(1 : 2)            :: binnedBadPixMapDim

    !**********************************
    ! Read in parameters.
    !**********************************
    sampleFactor = intParameter("samplefactor")
    timeFactor = intParameter("timefactor")
    set = stringParameter("set")

    thxSet = stringParameter("thxset")

    ! Read in the OM OSW Image file.

    inDataSetH = dataSet(set, modify)

    ! Determine file type from datamode in primary header.

    dataMode = stringAttribute(inDataSetH, "DATAMODE")

    !*************************************
    ! Check data mode okay
    !*************************************   
    call CheckValidDataMode(dataMode, "IMAGING", set)
     
    !*************************************************************
    ! Read in the science window keywords from the primary header.
    !*************************************************************
    binBPE = booleanAttribute(inDataSetH, "BINBPE")
    binAx1 = int32Attribute(inDataSetH, "BINAX1")
    binAx2 = int32Attribute(inDataSetH, "BINAX2")

    ! Get a handle on the modes binary table extension.

    inTableH = table(inDataSetH, "MODES")

    ! Get handles and pointers on the modes binary table extension columns and extract the window data.

    inColumnH(1) = column(inTableH, "WINDOWX0", read)
    windowX0P => int32Data(inColumnH(1))
    windowX0 = windowX0P(0)

    inColumnH(2) = column(inTableH, "WINDOWY0", read)
    windowY0P => int32Data(inColumnH(2))
    windowY0 = windowY0P(0)

    inColumnH(3) = column(inTableH, "WINDOWDX", read)
    windowDXP => int32Data(inColumnH(3))
    windowDX = windowDXP(0)

    inColumnH(4) = column(inTableH, "WINDOWDY", read)
    windowDYP => int32Data(inColumnH(4))
    windowDY = windowDYP(0)

    !**********************************************************
    ! 25.5.2001 - Need the missing-pixel value for the badpixelmap
    !**********************************************************
    missingPixelValue = real32Attribute(inDataSetH, "MISSPIX")
  
    !*************************************************
    ! Need to examine the image pixels for missing 
    ! pixels
    !************************************************* 
    imageArrayH = array(inDataSetH, "PRIMARY", read) 
    imageArrayP => real32Array2Data(imageArrayH)
 
    ! Release handles on data.
 
    do i = 1,  4
        call release(inColumnH(i))
    end do 
    call release(inTableH)

    !*****************************************************
    ! Read in the OM Tracking History Data Auxiliary file.
    !*****************************************************
    thxDataSetH = dataSet(thxSet, read)

    if (hasBlock(thxDataSetH, "ZERODRIFT")) then

       ! Issue warning

       write (errstr, *) "Missing OM Tracking History Data Auxiliary File. Using ZERODRIFT option"      
       call warning("ZERODRIFT")

       ! Set the state of the cal.

       call CAL_setState(OM, 1, 0)
       call CAL_getBadPixelMap(badPixMapP)

       !******************************
       ! Calculate the image binning.
       !******************************
       binX = 2 ** binAx1
       binY = 2 ** binAx2

       if (binBPE) then
          binX = binX * 2
          binY = binY * 2
       end if

       qualArrayDim(1) = windowDX / binX
       qualArrayDim(2) = windowDY / binY

       !**********************************************
       ! 19.10.2001 binnedBadPixMapDim added for
       ! array bounds checking
       !**********************************************
       binnedBadPixMapDim(1) = floor(2048.0 / binx) - 1
       binnedBadPixMapDim(2) = floor(2048.0 / biny) - 1

       allocate (binnedBadPixMap(0 : binnedBadPixMapDim(1), &
                                 0 : binnedBadPixMapDim(2)), stat=allocErr)
       call CheckAllocation(allocErr, "binnedBadPixelMap")

       allocate (windowedBadPixMap(0 : qualArrayDim(1) - 1, 0 : &
                                       qualArrayDim(2) - 1), stat=allocErr)
       call CheckAllocation(allocErr, "windowedBadPixelMap")

       !***********************  
       ! Bin the full image
       !***********************

       allocate(yIndex(0 : 2047), stat=allocErr)
       call CheckAllocation(allocErr, "yIndex")
       forall (j = 0 : 2047)  yIndex(j) =  min(binnedBadPixMapDim(2), floor(real(j) / real(biny)))

       forall(i = 0 : 2047)
          iPix = min(binnedBadPixMapDim(1), floor(real(i) / real(binx)))
          forall (j = 0 : 2047)
             jPix =  yIndex(j)
             binnedBadPixMap(iPix, jPix) = binnedBadPixMap(iPix, jPix) + badPixMapP(i, j)

          end forall
       end forall
       deallocate(yIndex, stat=allocErr)
       call CheckDeallocation(allocErr, "yIndex")



       !********************************************
       ! window the full image to the input image.
       !********************************************
       allocate(yIndex(0 : qualArrayDim(2) - 1), stat=allocErr)
       call CheckAllocation(allocErr, "yIndex")

       
       forall (j = 0 : (qualArrayDim(2) - 1)) yIndex(j) =  min(binnedBadPixMapDim(2), windowY0 / binY + j) 

       do i = 0, qualArrayDim(1) - 1
          iPix = min(binnedBadPixMapDim(1), windowX0 / binX + i) 
          do j= 0, qualArrayDim(2) - 1
              windowedBadPixMap(i, j) = binnedBadPixMap(iPix, yIndex(j))
  
          end do
       end do
       deallocate(yIndex, stat=allocErr)
       call CheckDeallocation(allocErr, "yIndex")


       qualArrayH = addArray(inDataSetH, "QUALITY", Integer8, qualArrayDim)
       qualArrayP => int8Array2Data(qualArrayH)

       ! Convert to a boolean bad pixel map.

       !
       !--------------------------------------------------------------------------------------------
       !
       ! AMK - 24/08/200
       !
       ! Changed initialisation below to correct integer valuews and NOT reals
       !
       !      qualArrayP = 0.0
       !      where (windowedBadPixMap /= 0) qualArrayP = 1.0

       qualArrayP = 0


       !**************************************************
       ! Check the image for missing pixels or negative
       ! values- if any set corresponding quality-map pixel
       ! to 1
       !**************************************************
       do i = 0, size(imageArrayP, 1) - 1
              
           do j = 0, size(imageArrayP, 2) - 1
      
               if (abs(imageArrayP(i, j) - missingPixelValue) .lt. 0.5) then
                   qualArrayP(i, j) = 1
                   imageArrayP(i, j) = 0.0 
               else
                   if (imageArrayP(i, j) .lt. 0.0) then

                       qualArrayP(i, j) = 1
                       imageArrayP(i, j) = 0.0 
                   end if
               end if
           end do
       end do


       where (windowedBadPixMap .ne. 0) qualArrayP = 1
     

       ! Release the cal memory.

       call cal_releaseMemory(badPixMapP)

       call release(qualArrayH)

    else
       ! Determine file type from datamode in primary header.

       dataMode = stringAttribute(thxDataSetH, "DATAMODE")

       !********************************
       ! Check data mode okay
       !******************************** 
       call CheckValidDataMode(dataMode, "TRACKING", thxSet)
     
       ! Get a handle on the tracking history binary table extension.

       thxTableH = table(thxDataSetH, "OMTHX1")

       ! Get the number of frames in the tracking history binary table extension.

       nFrames = numberOfRows(thxTableH)

       ! Allocate space for the tracking frames.

       allocate (tf(0 : nFrames), stat=allocErr)
       call CheckAllocation(allocErr, "tf")


       ! Get handles and pointers on the tracking history binary table extension columns and extract the tracking data.

       thxColumnH(1) = column(thxTableH, "DX", read)
       dXP => int32Data(thxColumnH(1))
       tf(0)%dX = 0.0
       tf(1 : nFrames)%dX = real(dXP / 1000.0)

       thxColumnH(2) = column(thxTableH, "DY", read)
       dYP => int32Data(thxColumnH(2))
       tf(0)%dY = 0.0
       tf(1 : nFrames)%dY = real(dYP / 1000.0)

       thxColumnH(3) = column(thxTableH, "QUALITY", read)
       qualityP => int32Data(thxColumnH(3))
       tf(0)%quality = 0.0
       tf(1 : nFrames)%quality = real(qualityP / 1000000.0)

       ! Release handles on data.

       do i = 1,  3 
           call release(thxColumnH(i))
       end do

       call release(thxTableH)
       call release(thxDataSetH)

       ! Set the state of the cal.

       call CAL_setState(OM, 1, 0)

       ! Get the bad pixel map.
       call CAL_getBadPixelMap(badPixMapP)

          ! Allocate space for the boolean bad pixel map.

       allocate (badPixMap(0 : 2047, 0 : 2047), stat=allocErr)
       call CheckAllocation(allocErr, "badPixMap")


       ! Convert to a boolean bad pixel map.

       badPixMap = .false.
       where (badPixMapP /= 0) badPixMap = .true.
  
       ! Release the cal memory.

       call cal_releaseMemory(badPixMapP)


       ! Calculate the maximum drift in x and y.
       ! We add 0.5 to the drifts so that positions are calculated with respect to the centre of the
       ! pixel rather than the bottom left hand corner.

       maxDX = 0
       maxDY = 0
       forall (i = 1 : nFrames)
          maxDX = max(maxDX, ceiling(0.5 + abs(tf(i)%dX)))
          maxDY = max(maxDY, ceiling(0.5 + abs(tf(i)%dY)))
       end forall

       ! Allocate space for the loaclised bad pixel map.

       cpxMin = windowX0 - maxDX
       cpxMax = windowX0 + windowDX - 1 + maxDX
       cpyMin = windowY0 - maxDY
       cpyMax = windowY0 + windowDY - 1 + maxDY
       allocate (lBadPixMap(cpxMin:cpxMax, cpyMin:cpyMax), stat=allocErr)
       call CheckAllocation(allocErr, "lBadPixMap")


       ! Fill localised bad pixel map.

       lBadPixMap = .false.
       forall (i = max(0, cpxMin) : min(2047, cpxMax))
           forall (j = max(0, cpyMin) : min(2047, cpyMax)) lBadPixMap(i, j) = badPixMap(i, j)
       end forall

       ! Free up memory.

       deallocate(badPixMap, stat=allocErr)
       call CheckDeallocation(allocErr, "badPixMap")


       ! Allocate space for the oversampled bad pixel map.

       oMaxDX = maxDX * sampleFactor
       oMaxDY = maxDY * sampleFactor
       opxMin = 0 - oMaxDX
       opxMax = ((cpxMax - cpxMin - (2 * maxDX)) * sampleFactor) + oMaxDX
       opyMin = 0 - oMaxDY
       opyMax = ((cpyMax - cpyMin - (2 * maxDY)) * sampleFactor) + oMaxDY
       allocate (oBadPixMap(opxMin:opxMax, opyMin:opyMax), stat=allocErr)
       call CheckAllocation(allocErr, "oBadPixmap")


       !**********************************
       ! Oversample the bad pixel map.
       !*********************************
       oBadPixMap = .false.

       allocate(yIndex(opyMin : opyMax))
       forall (j = opyMin : opyMax) yIndex(j) =  floor((j + oMaxDY) / real(sampleFactor)) + cpyMin

       forall(i = opxMin : opxMax)           
          shiftX = floor((i + oMaxDX) / real(sampleFactor))
          x = cpxMin + shiftX
          forall(j = opyMin : opyMax)      
              oBadPixMap(i, j) = lBadPixMap(x, yIndex(j))

          end forall
       end forall
      

       ! Free up memory.

       deallocate(lBadPixMap, stat=allocErr)
       call CheckDeallocation(allocErr, "lBadPixMap")


       ! Allocate space for the oversampled quality array.

       opxMin = 0
       opxMax = (cpxMax - cpxMin - (2 * maxDX)) * sampleFactor
       opyMin = 0
       opyMax = (cpyMax - cpyMin - (2 * maxDY)) * sampleFactor

       allocate (oQualArray(opxMin:opxMax, opyMin:opyMax), stat=allocErr)
       call CheckAllocation(allocErr, "qQualArray")

       !****************************************************************************************** 
       ! Loop over the tracking frames and construct an oversampled tracking shifted quality array.
       ! When we apply the shifts, we must centre the star in the pixel, so we add 0.5 to x
       ! and y or in this case to dx and dy.
       !******************************************************************************************

       oQualArray = .false.
       do i = 1, nFrames
          rShiftX = (tf(i)%dx - tf(i - 1)%dx) * sampleFactor
          rShiftY = (tf(i)%dy - tf(i - 1)%dy) * sampleFactor
          xConstant =  (0.5 + tf(i - 1)%dx) * sampleFactor
          yConstant =  (0.5 + tf(i - 1)%dy) * sampleFactor
          do j = 0, timeFactor
    
             iShiftX = nint(xConstant + (rShiftX * (j / real(timeFactor))))
             iShiftY = nint(yConstant + (rShiftY * (j / real(timeFactor))))
             iXMin = opxMin - iShiftX
             iXMax = opxMax - iShiftX
             iYMin = opyMin - iShiftY
             iYMax = opyMax - iShiftY
             oQualArray = oQualArray .or. oBadPixMap(iXMin : iXMax, iYMin : iYMax)
          end do
       end do

       ! Free up memory.

       deallocate(oBadPixMap, stat=allocErr)
       call CheckDeallocation(allocErr, "oBadPixMap")

       ! Allocate space for the quality array.

       cpxMin = windowX0
       cpxMax = windowX0 + windowDX - 1
       cpyMin = windowY0
       cpyMax = windowY0 + windowDY - 1

       allocate (qualArray(cpxMin:cpxMax, cpyMin:cpyMax), stat=allocErr)
       call CheckAllocation(allocErr, "qualArray")

       !**************************
       ! Fill the quality array.
       !**************************

       qualArray = .false.
       opxMin = 0
       opxMax = (cpxMax - cpxMin) * sampleFactor
       opyMin = 0
       opyMax = (cpyMax - cpyMin) * sampleFactor


       forall (j = opyMin : opyMax) yIndex(j) = floor(j / real(sampleFactor)) + cpyMin 

       forall( i = opxMin : opxMax)
          shiftX = floor(i / real(sampleFactor))
          x = cpxMin + shiftX
          forall (j = opyMin : opyMax)
              qualArray(x, yIndex(j)) = oQualArray(i, j)
          end forall
       end forall

       deallocate(yIndex, stat=allocErr)
       call CheckDeallocation(allocErr, "yIndex")


       ! Free up memory.

       deallocate(oQualArray, stat=allocErr)
       call CheckDeallocation(allocErr, "oQualArray")


       ! Calculate the image binning.

       binX = 2**binAx1
       binY = 2**binAx2

       if (binBPE) then
          binX = binX * 2
          binY = binY * 2
       end if

       qualArrayDim = shape(qualArray)

       qualArrayDim(1) = qualArrayDim(1) / binX
       qualArrayDim(2) = qualArrayDim(2) / binY

       ! Add the quality extension to the image file.

       qualArrayH = addArray(inDataSetH, "QUALITY", Integer8, qualArrayDim)
       qualArrayP => int8Array2Data(qualArrayH)

       !********************************
       ! Bin the quality array.
       !********************************
       qualArrayP = 0
   
       allocate(yIndex(cpyMin : cpyMax), stat=allocErr)
       call CheckAllocation(allocErr, "yIndex")

       iMax =  qualArrayDim(1) - 1
       jMax =  qualArrayDim(2) - 1

       forall (j = cpyMin : cpyMax) 
           shiftY = j - cpyMin
           y = floor(shiftY / real(binY))
           yIndex(j) = min(y, jMax)    
       end forall 
    
       do i = cpxMin, cpxMax
          shiftX = i - cpxMin
          x = floor(shiftX / real(binX))
          x = min(x, iMax)       ! Added 12.7.2000 to prevent array bound error
          do j = cpyMin, cpyMax
      
             if (qualArray(i, j)) qualArrayP(x, yIndex(j)) = 1
             
          end do
       end do


       deallocate(yIndex, stat=allocErr)
       call CheckDeallocation(allocErr, "yIndex")


       ! Free up memory.

       deallocate(qualArray, stat = allocErr)
       call CheckDeallocation(allocErr, "qualArray")


       ! Release handles on data.

       call release(qualArrayH)

    endif
    call release(imageArrayH)
    call release(inDataSetH)

    !////////////////////////////////////////////////////////////////
    !
    ! AMK - Tidy up allocated memomry
    !
    !///////////////////////////////////////////////////////////////

    if(allocated(binnedBadPixMap)) then
       deallocate(binnedBadPixMap, stat = istat)
       call CheckDeallocation(istat, "binnedBadPixMap")

    end if

    if(allocated(windowedBadPixMap)) then
       deallocate(windowedBadPixMap,stat=istat)
       call CheckDeallocation(istat, "windowedBadPixmap")

    end if

    if(allocated(tf)) then
       deallocate(tf, stat = istat)
       call CheckDeallocation(istat, "tf")

    end if

    if(allocated(badPixMap)) then
       deallocate(badPixMap,stat = istat)
       call CheckDeallocation(istat, "badPixMap")
    end if

    if(allocated(lBadPixMap)) then
       deallocate(lBadPixMap,stat = istat)
       call CheckDeallocation(istat, "lBadPixMap")

    end if

    if(allocated(oBadPixMap)) then
       deallocate(oBadPixMap,stat = istat)
       call CheckDeallocation(istat, "oBadPixMap")

    end if

    if(allocated(oQualArray)) then
       deallocate(oQualArray,stat = istat)
       call CheckDeallocation(istat, "qualArray")

    end if

    if(allocated(qualArray)) then
       deallocate(qualArray,stat = istat)
       call CheckDeallocation(istat, "qualArray")

    end if

end subroutine omcosflag
!***************************************************
! CheckAllocation
!***************************************************
subroutine CheckAllocation(status, string)
    integer, intent(in) ::  status
    character(len=*), intent(in) :: string
    character(len=100) :: comment
   
    if(status .ne. 0) then
       write(comment, '(a, a)') "Could not allocate memory for : ", string 
       write (errstr, *) comment
       call fatal("allocErr") 
    end if

end subroutine CheckAllocation
!***************************************************
! CheckDeallocation
!***************************************************
subroutine CheckDeallocation(status, string)
    integer, intent(in) ::  status
    character(len=*), intent(in) :: string
    character(len=100) :: comment
   
    if(status .ne. 0) then
       write(comment, '(a, a)') "Failed to release memory for : ", trim(string) 
       write (errstr, *) comment
       call warning("allocErr", comment) 
    end if

end subroutine CheckDeallocation
!**********************************
! Subroutine CheckValidDataMode
!**********************************
subroutine CheckValidDataMode(dataMode, validMode, fileName)
    character(len=*), intent(in) :: dataMode
    character(len=*), intent(in) :: validMode
    character(len=*), intent(in) :: fileName


    if (trim(dataMode) .ne. trim(validMode)) then
 
        write (errstr, *) "Invalid datamode ", dataMode, " in set ", fileName
        call fatal("badFileMode")
    end if

end subroutine CheckValidDataMode

end module omcosflag_mod













